#class to use dhan api

import logging
import requests
from json import loads as json_loads, dumps as json_dumps
from pathlib import Path
from webbrowser import open as web_open
from datetime import datetime, timedelta


class dhanhq:

    #declaring constants
    NSE= 'NSE_EQ'
    BSE= 'BSE_EQ'
    CUR= 'NSE_CURRENCY'
    MCX= 'MCX_COMM'
    FNO= 'NSE_FNO'
    NSE_FNO = 'NSE_FNO'
    BSE_FNO = 'BSE_FNO'
    BUY= B= 'BUY'
    SELL= S= 'SELL'
    CNC= 'CNC'
    INTRA= "INTRADAY"
    SL= "STOP_LOSS"
    SLM= "STOP_LOSS_MARKET"
    MARGIN= 'MARGIN'
    CO= 'CO'
    BO= 'BO'
    MTF= 'MTF'
    LIMIT= 'LIMIT'
    MARKET= 'MARKET'
    DAY= 'DAY'
    IOC= 'IOC'
    GTC= 'GTC'
    GTD= 'GTD'
    EQ= 'EQ'
    def __init__(self,client_id,access_token):
        try:
            self.client_id= str(client_id)
            self.access_token= access_token
            self.base_url= 'https://api.dhan.co'
            self.timeout= 60 #used for http requests
            self.header= {
                'access-token': access_token,
                'content-type': 'application/json',
            }
            requests.packages.urllib3.util.connection.HAS_IPV6 = False
            self.session= requests.Session()
        except Exception as e:
            logging.error('Exception in dhanhq>>init : %s',e)

    def _parse_response(self,response):
        try:
            status= 'failure'
            remarks=''
            data=''
            python_response= json_loads(response.content)
            if response.status_code==200:
                status= 'success'
                remarks=''
                data= python_response
            else:
                error_code=python_response['internalErrorCode']
                error_logs= python_response['internalErrorMessage']
                remarks= {
                    'error_code':error_code,
                    'message':error_logs
                }
        except Exception as e:
            logging.warning('Exception found in dhanhq>>find_error_code: %s',e)
            status= 'failure'
            remarks= str(e)
        return {
            'status':status,
            'remarks':remarks,
            'data':data,
        }

    def get_order_list(self):
        """Retrieve a list of all orders requested in a day with their last updated status."""
        try:
            url= self.base_url+'/orders'
            response= self.session.get(url,headers=self.header,timeout= self.timeout)
            return self._parse_response(response)
        except Exception as e:
            logging.error('Exception in dhanhq>>get_order_list : %s',e)
            return {
                'status':'failure',
                'remarks':f'Exception in dhanhq>>get_order_list : {e}',
                'data':'',
            }

    def get_order_by_id(self,order_id):
        """Retrieve the details and status of an order from the orderbook placed during the day."""
        try:
            url= self.base_url+f'/orders/{order_id}'
            response= self.session.get(url,headers=self.header,timeout= self.timeout)
            return self._parse_response(response)
        except Exception as e:
            logging.error('Exception in dhanhq>>get_order_by_id : %s',e)
            error= str(e)
            return {
                'status':'failure',
                'remarks':f'Exception in dhanhq>>get_order_by_id : {e}',
                'data':'',
            }

    def modify_order(self,order_id,order_type,leg_name,quantity,price,trigger_price,disclosed_quantity,validity):
        """Modify pending order in orderbook. The variables that can be modified are price, quantity, order type & validity."""
        try:
            url= self.base_url+f'/orders/{order_id}'
            payload={
                    "dhanClientId": self.client_id,
                    "orderId": str(order_id),
                    "orderType": order_type,
                    "legName": leg_name,
                    "quantity": quantity,
                    "price": price,
                    "disclosedQuantity": disclosed_quantity,
                    "triggerPrice": trigger_price,
                    "validity": validity
            }
            payload= json_dumps(payload)
            response= self.session.put(url,headers=self.header,timeout= self.timeout, data=payload)
            return self._parse_response(response)
        except Exception as e:
            logging.error('Exception in dhanhq>>modify_order: %s',e)
            error= str(e)
            return {
                'status':'failure',
                'remarks':str(e),
                'data':'',
            }

    def cancel_order(self,order_id):
        """Cancel a pending order in the orderbook using the order id of an order."""
        try:
            url= self.base_url+f'/orders/{order_id}'
            response= self.session.delete(url,headers=self.header,timeout= self.timeout)
            return self._parse_response(response)
        except Exception as e:
            logging.error('Exception in dhanhq>>cancel_order: %s',e)
            return {
                'status':'failure',
                'remarks':str(e),
                'data':'',
            }

    def place_order(self,security_id,exchange_segment,transaction_type,quantity,\
        order_type,product_type,price,trigger_price=0,disclosed_quantity=0,\
        after_market_order=False,validity='DAY',amo_time='OPEN',\
        bo_profit_value=None,bo_stop_loss_Value=None,\
        drv_expiry_date=None,drv_options_type=None,drv_strike_price=None,tag=None):
        """Place new Orders"""
        #place order in Dhan account
        #security_id(str),exchange_segment(str),transaction_type(str),
        #quantity(int),order_type(str),validity(str),product_type(str),
        #price(float),trigger_price(float),disclosed_quantity(int),
        #after_market_order(Boolean),amo_time(str),tag(str)

        try:
            url= self.base_url+'/orders'
            payload={
                    "dhanClientId": self.client_id,
                    "transactionType": transaction_type.upper(),
                    "exchangeSegment": exchange_segment.upper(),
                    "productType": product_type.upper(),
                    "orderType": order_type.upper(),
                    "validity": validity.upper(),
                    "securityId": security_id,
                    "quantity": int(quantity),
                    "disclosedQuantity": int(disclosed_quantity),
                    "price": float(price),
                    "afterMarketOrder": after_market_order,
                    "boProfitValue":bo_profit_value,
                    "boStopLossValue":bo_stop_loss_Value,
                    "drvExpiryDate":drv_expiry_date,
                    "drvOptionType":drv_options_type,
                    "drvStrikePrice":drv_strike_price
                }
            if tag!=None and tag!='':
                payload["correlationId"] = tag
            if after_market_order== True:
                if amo_time in ['OPEN','OPEN_30','OPEN_60']:
                    payload['amoTime'] = amo_time
                else:
                    raise Exception("amo_time value must be ['OPEN','OPEN_30','OPEN_60']")
            if trigger_price>0:
                payload["triggerPrice"]= float(trigger_price)
            elif trigger_price==0:
                payload["triggerPrice"]= 0.0

            payload= json_dumps(payload)
            response= self.session.post(url,data=payload,headers=self.header,timeout=self.timeout)
            return self._parse_response(response)
        except Exception as e:
            logging.error('Exception in dhanhq>>place_order: %s',e)
            return {
                'status':'failure',
                'remarks':str(e),
                'data':'',
            }

    def get_order_by_corelationID(self,corelationID):
        """Retrieves the order status using a field called correlation id, Provided by API consumer during order placement."""
        try:
            url= self.base_url+f'/orders/external/{corelationID}'
            response= self.session.get(url,headers=self.header,timeout=self.timeout)
            return self._parse_response(response)
        except Exception as e:
            logging.error('Exception in dhanhq>>get_order_by_corelationID: %s',e)
            return {
                'status':'failure',
                'remarks':str(e),
                'data':'',
                'error_logs':''
            }

    def get_positions(self):
        """Retrieve a list of all open positions for the day. This includes all F&O carryforward positions as well."""
        try:
            url= self.base_url+f'/positions'
            response= self.session.get(url,headers=self.header,timeout=self.timeout)
            return self._parse_response(response)
        except Exception as e:
            logging.error('Exception in dhanhq>>get_positions: %s',e)
            return {
                'status':'failure',
                'remarks':str(e),
                'data':'',
            }

    def get_holdings(self):
        """Retrieve all holdings bought/sold in previous trading sessions. All T1 and delivered quantities can be fetched."""
        try:
            url= self.base_url+f'/holdings'
            response= self.session.get(url,headers=self.header,timeout=self.timeout)
            return self._parse_response(response)
        except Exception as e:
            logging.error('Exception in dhanhq>>get_holdings: %s',e)
            return {
                'status':'failure',
                'remarks':str(e),
                'data':'',
            }

    def intraday_daily_minute_charts(self,security_id,exchange_segment,instrument_type):
        """Retrieve OHLC & Volume of 1 minute candle for desired instrument for current day. This data available for all segments including futures & options."""
        try:
            url= self.base_url+f'/charts/intraday'
            payload= {
                'securityId':security_id,
                'exchangeSegment':exchange_segment,
                'instrument':instrument_type
            }
            payload= json_dumps(payload)
            response= self.session.post(url,headers=self.header,timeout=self.timeout,data=payload)
            return self._parse_response(response)
        except Exception as e:
            logging.error('Exception in dhanhq>>intraday_daily_minute_charts: %s',e)
            return {
                'status':'failure',
                'remarks':str(e),
                'data':'',
            }
        
    # Update nomenclature for v1.2.4
    def intraday_minute_data(self,security_id,exchange_segment,instrument_type):
        """Retrieve OHLC & Volume of 1 minute candle for desired instrument for current day. This data available for all segments including futures & options."""
        try:
            url= self.base_url+f'/charts/intraday'
            payload= {
                'securityId':security_id,
                'exchangeSegment':exchange_segment,
                'instrument':instrument_type
            }
            payload= json_dumps(payload)
            response= self.session.post(url,headers=self.header,timeout=self.timeout,data=payload)
            return self._parse_response(response)
        except Exception as e:
            logging.error('Exception in dhanhq>>intraday_daily_minute_charts: %s',e)
            return {
                'status':'failure',
                'remarks':str(e),
                'data':'',
            }

    def historical_minute_charts(self,symbol,exchange_segment,instrument_type,expiry_code,from_date,to_date):
        """Retrieve OHLC & Volume of daily candle for desired instrument. The data for any scrip is available back upto the date of its inception."""
        try:
            url= self.base_url+f'/charts/historical'
            payload= {
                    "symbol": symbol,
                    "exchangeSegment": exchange_segment,
                    "instrument": instrument_type,
                    "expiryCode": expiry_code,
                    "fromDate": from_date,
                    "toDate": to_date
                    }
            payload= json_dumps(payload)
            response= self.session.post(url,headers=self.header,timeout=self.timeout,data=payload)
            return self._parse_response(response)
        except Exception as e:
            logging.error('Exception in dhanhq>>intraday_history_minute_charts: %s',e)
            return {
                'status':'failure',
                'remarks':str(e),
                'data':'',
            }
    
    # Update nomenclature for v1.2.4
    def historical_daily_data(self,symbol,exchange_segment,instrument_type,expiry_code,from_date,to_date):
        """Retrieve OHLC & Volume of daily candle for desired instrument. The data for any scrip is available back upto the date of its inception."""
        try:
            url= self.base_url+f'/charts/historical'
            payload= {
                    "symbol": symbol,
                    "exchangeSegment": exchange_segment,
                    "instrument": instrument_type,
                    "expiryCode": expiry_code,
                    "fromDate": from_date,
                    "toDate": to_date
                    }
            payload= json_dumps(payload)
            response= self.session.post(url,headers=self.header,timeout=self.timeout,data=payload)
            return self._parse_response(response)
        except Exception as e:
            logging.error('Exception in dhanhq>>intraday_history_minute_charts: %s',e)
            return {
                'status':'failure',
                'remarks':str(e),
                'data':'',
            }

    def get_trade_book(self,order_id=None):
        """Retrieve a list of all trades executed in a day."""
        try:
            if order_id==None:
                url= self.base_url+f'/trades'
            else:
                url= self.base_url+f'/trades/{order_id}'
            response= self.session.get(url,headers=self.header,timeout=self.timeout)
            return self._parse_response(response)
        except Exception as e:
            logging.error('Exception in dhanhq>>get_trade_book: %s',e)
            return {
                'status':'failure',
                'remarks':str(e),
                'data':'',
            }

    def get_trade_history(self,from_date,to_date,page_number=0):
        """Retrieve the trade history Often during partial trades or Bracket/ Cover Orders, traders get confused in reading trade from tradebook. The response of this API will include all the trades generated for a particular order id."""
        try:
            url= self.base_url+f'/tradeHistory/{from_date}/{to_date}/{page_number}'
            response= self.session.get(url,headers=self.header,timeout=self.timeout)
            return self._parse_response(response)
        except Exception as e:
            logging.error('Exception in dhanhq>>get_trade_history: %s',e)
            return {
                'status':'failure',
                'remarks':str(e),
                'data':'',
            }

    def get_fund_limits(self):
        """Get all information of your trading account like balance, margin utilised, collateral, etc."""
        try:
            url= self.base_url+f'/fundlimit'
            response= self.session.get(url,headers=self.header,timeout=self.timeout)
            return self._parse_response(response)
        except Exception as e:
            logging.error('Exception in dhanhq>>get_fund_limits: %s',e)
            return {
                'status':'failure',
                'remarks':str(e),
                'data':'',
            }

    def place_slice_order(self,security_id,exchange_segment,transaction_type,quantity,\
        order_type,product_type,price,trigger_price=0,disclosed_quantity=0,\
        after_market_order=False,validity='DAY',amo_time='OPEN',\
        bo_profit_value=None,bo_stop_loss_Value=None,\
        drv_expiry_date=None,drv_options_type=None,drv_strike_price=None,tag=None):
        """Place new Slice Orders"""
        #place order in Dhan account
        #security_id(str),exchange_segment(str),transaction_type(str),
        #quantity(int),order_type(str),validity(str),product_type(str),
        #price(float),trigger_price(float),disclosed_quantity(int),
        #after_market_order(Boolean),amo_time(str),tag(str)

        try:
            url= self.base_url+'/orders/slicing'
            payload={
                    "dhanClientId": self.client_id,
                    "transactionType": transaction_type.upper(),
                    "exchangeSegment": exchange_segment.upper(),
                    "productType": product_type.upper(),
                    "orderType": order_type.upper(),
                    "validity": validity.upper(),
                    "securityId": security_id,
                    "quantity": int(quantity),
                    "disclosedQuantity": int(disclosed_quantity),
                    "price": float(price),
                    "afterMarketOrder": after_market_order,
                    "boProfitValue":bo_profit_value,
                    "boStopLossValue":bo_stop_loss_Value,
                    "drvExpiryDate":drv_expiry_date,
                    "drvOptionType":drv_options_type,
                    "drvStrikePrice":drv_strike_price
                }
            if tag!=None and tag!='':
                payload["correlationId"] = tag
            if after_market_order== True:
                if amo_time in ['OPEN','OPEN_30','OPEN_60']:
                    payload['amoTime'] = amo_time
                else:
                    raise Exception("amo_time value must be ['OPEN','OPEN_30','OPEN_60']")
            if trigger_price>0:
                payload["triggerPrice"]= float(trigger_price)
            elif trigger_price==0:
                payload["triggerPrice"]= 0.0

            payload= json_dumps(payload)
            response= self.session.post(url,data=payload,headers=self.header,timeout=self.timeout)
            return self._parse_response(response)
        except Exception as e:
            logging.error('Exception in dhanhq>>place_order: %s',e)
            return {
                'status':'failure',
                'remarks':str(e),
                'data':'',
            }

    def generate_tpin(self):
        """Generate T-Pin on registered mobile number"""
        try:
            url= self.base_url+'/edis/tpin'
            response= self.session.get(url,headers=self.header,timeout=self.timeout)
            if response.status_code==202:
                return {
                    'status':'success',
                    'remarks':'Otp sent',
                    'data':''
                }
            else:
                return {
                    'status':'failure',
                    'remarks':'status code :'+response.status_code,
                    'data':'',
                }
        except Exception as e:
            logging.error('Exception in dhanhq>>generate_tpin : %s',e)
            return {
                'status':'failure',
                'remarks':str(e),
                'data':'',
            }

    def open_browser_for_tpin(self,isin,qty,exchange,segment='EQ',bulk= False):
        """Opens default web browser for enter tpin"""
        try:
            url= self.base_url+'/edis/form'
            data= {
                    "isin": isin,
                    "qty": qty,
                    "exchange": exchange,
                    "segment": segment,
                    "bulk": bulk
                }
            data= json_dumps(data)
            response= self.session.post(url,headers=self.header,data=data,timeout=self.timeout)
            data= json_loads(response.content)
            form_html= data['edisFormHtml']
            form_html= form_html.replace('\\','')
            with open("temp_form.html", "w") as f:
                f.write(form_html)
            filename = f'file:\\\{Path.cwd()}\\temp_form.html'
            web_open(filename)
            return self._parse_response(response)
        except Exception as e:
            logging.error('Exception in dhanhq>>open_browser_for_tpin : %s',e)
            return {
                'status':'failure',
                'remarks':str(e),
                'data':'',
            }

    def edis_inquiry(self,isin):
        """Inquires about provided isin"""
        try:
            url= f'{self.base_url}/edis/inquire/{isin}'
            response= self.session.get(url,headers=self.header,timeout=self.timeout)
            return self._parse_response(response)
        except Exception as e:
            logging.error('Exception in dhanhq>>edis_inquiry : %s',e)
            return {
                'status':'failure',
                'remarks':str(e),
                'data':'',
            }

    def convert_to_date_time(self,JulianDate):
        """Convert julian date to python datetime object"""
        Dt1980= datetime(year=1980,month=1,day=1,hour=5,minute=30)
        DtObj= Dt1980+ timedelta(seconds=JulianDate)
        return DtObj



"""
    The marketfeed class is designed to facilitate asynchronous communication with the DhanHQ API via WebSocket. 
    It enables users to subscribe to market data for a list of instruments and receive real-time updates.

    :copyright: (c) 2024 by Dhan.
    :license: see LICENSE for details.
"""

import websockets
import asyncio
import json
import struct
from datetime import datetime

# Constants
"""WebSocket URL for DhanHQ Live Market Feed"""
WSS_URL = 'wss://api-feed.dhan.co'

"""Constants for Exchange Segment"""
IDX = 0
NSE = 1
NSE_FNO = 2
NSE_CURR = 3
BSE = 4
MCX = 5
BSE_CURR = 7
BSE_FNO = 8

"""Constants for Request Code"""
Ticker = 15
Quote = 17
Depth = 19


class DhanSDKHelper:
    def __init__(self, sdk_instance):
        self.sdk_instance = sdk_instance

    """Callback function executed when the WebSocket connection is established."""
    async def on_connection_established(self, websocket):
        if self.sdk_instance.on_connect:
            await self.sdk_instance.on_connect(self.sdk_instance)

    """Callback function executed when a message is received."""
    async def on_message_received(self, response):
        if self.sdk_instance.on_message:
            await self.sdk_instance.on_message(self.sdk_instance, response)

    """Callback function executed when the WebSocket connection is closed."""
    async def on_close(self, websocket, close_status=None, close_msg=None):
        print(f"WebSocket closed with status {close_status}: {close_msg}")
        await websocket.close()
        self.ws = None



class DhanFeed:
    def __init__(self, client_id, access_token, 
                 instruments, subscription_code, 
                 on_connect=None, 
                 on_message=None, 
                 on_close=None):
        """Initializes the DhanSDK instance with user credentials, instruments to subscribe, and callback functions."""
        
        self.client_id = client_id
        self.access_token = access_token
        self.instruments = instruments
        self.subscription_code = subscription_code
        self.data = ""
        self._is_first_connect = True
        self.ws = None
        self.on_ticks = None
        self.on_message = on_message
        self.on_close = on_close
        self.on_connect = on_connect
        self.loop = asyncio.get_event_loop()

        if self.on_connect:
            self.connect()

    def run_forever(self):
        """Starts the WebSocket connection and runs the event loop."""
        self.loop.run_until_complete(self.connect())

    async def connect(self):
        """Initiates the connection to the Websockets"""
        if not self.ws or self.ws.closed:
            self.ws = await websockets.connect(WSS_URL)
            helper = DhanSDKHelper(self)
            await helper.on_connection_established(self.ws)
            await self.authorize()
            await self.subscribe_instruments()

            # Handling incoming messages in a loop to keep the connection open
            while True:
                try:
                    response = await self.ws.recv()
                    self.data = self.process_data(response)
                    await helper.on_message_received(self.data)
                except websockets.exceptions.ConnectionClosed:
                    print("Connection has been closed")
                    break

    async def authorize(self):
        """Establishes the WebSocket connection and authorizes the user"""
        try:
            helper = DhanSDKHelper(self)
            await helper.on_connection_established(self.ws)

            # Authorization packet creation
            api_access_token = self.access_token.encode('utf-8')
            api_access_token = self.pad_with_zeros(api_access_token, 500)
            authentication_type = "2P".encode('utf-8')
            payload = api_access_token + authentication_type
            
            feed_request_code = 11
            message_length = 83 + len(api_access_token) + len(authentication_type)
            client_id = self.client_id.encode('utf-8')
            client_id = self.pad_with_zeros(client_id, 30)
            dhan_auth = b"\0" * 50
            header = struct.pack('<bH30s50s', feed_request_code, message_length, client_id, dhan_auth)

            authorization_packet = header + payload

            # Send authorization packet
            await self.ws.send(authorization_packet)
            self.is_authorized = True

        except Exception as e:
            print(f"Authorization failed: {e}")
            self.is_authorized = False
        
    async def subscribe_instruments(self):
        """Subscribe Instruments on the Open Websocket"""
        if not self.is_authorized:
            print("Not authorized. Please authorize first.")
            return

        # Subscription packet creation
        subscription_packet = self.create_subscription_packet(self.instruments, self.subscription_code)
        await self.ws.send(subscription_packet)

    def process_data(self, data):
        """Read binary data and initiate processing in received format"""
        first_byte = struct.unpack('<B', data[0:1])[0]
        if first_byte == 2:
            return self.process_ticker(data)
        elif first_byte == 3:
            return self.process_market_depth(data)
        elif first_byte == 4:
            return self.process_quote(data)
        elif first_byte == 5:
            return self.process_oi(data)
        elif first_byte == 6:
            return self.process_prev_close(data)
        elif first_byte == 7:
            return self.process_status(data)
        elif first_byte == 50:
            return self.server_disconnection(data)

    def process_ticker(self, data):
        """Parse and process Ticker Data"""
        unpack_ticker = [struct.unpack('<BHBIfI', data[0:16])]
        # Assign unpacked data to fields
        ticker_data = {
            "type" : 'Ticker Data',
            "exchange_segment" : unpack_ticker[0][2],
            "security_id" : unpack_ticker[0][3],
            "LTP" : "{:.2f}".format(unpack_ticker[0][4]),
            "LTT" : self.utc_time(unpack_ticker[0][5])
        } 
        return ticker_data
    
    def process_prev_close(self, data):
        """Parse and process Previous Day Data"""
        unpack_pclose = [struct.unpack('<BHBIfI', data[0:16])]
        # Assign unpacked data to fields
        prev_close = {
            "type" : 'Previous Close',
            "exchange_segment" : unpack_pclose[0][2],
            "security_id" : unpack_pclose[0][3],
            "prev_close" : "{:.2f}".format(unpack_pclose[0][4]),
            "prev_OI" : unpack_pclose[0][5]
        } 
        return prev_close

    def process_market_depth(self, data):
        """Parse and process Market Depth Data"""
        market_data = [struct.unpack('<BHBIf100s', data[0:112])]
        market_data_list = list(market_data[0][0:5])
        market_depth_binary = market_data[0][5]
        packet_format = '<IIHHff'

        for i in range(5):
            start_idx = i * struct.calcsize(packet_format)
            end_idx = start_idx + struct.calcsize(packet_format)
            current_packet = struct.unpack(packet_format, market_depth_binary[start_idx:end_idx])
            market_data_list.append(current_packet)
        exchange_segment = market_data_list[2]
        sec_id = market_data_list[3]
        ltp = market_data_list[4]
        depth1 = market_data_list[5]
        depth2 = market_data_list[6]
        depth3 = market_data_list[7]
        depth4 = market_data_list[8]
        depth5 = market_data_list[9]
        depth = [{"bid_quantity": depth1[0], "ask_quantity": depth1[1], "bid_orders": depth1[2], "ask_orders": depth1[3],
                  "bid_price": depth1[4], "ask_price": depth1[5]},
                 {"bid_quantity": depth2[0], "ask_quantity": depth2[1], "bid_orders": depth2[2], "ask_orders": depth2[3],
                  "bid_price": depth2[4], "ask_price": depth2[5]},
                 {"bid_quantity": depth3[0], "ask_quantity": depth3[1], "bid_orders": depth3[2], "ask_orders": depth3[3],
                  "bid_price": depth3[4], "ask_price": depth3[5]},
                 {"bid_quantity": depth4[0], "ask_quantity": depth4[1], "bid_orders": depth4[2], "ask_orders": depth4[3],
                  "bid_price": depth4[4], "ask_price": depth4[5]},
                 {"bid_quantity": depth5[0], "ask_quantity": depth5[1], "bid_orders": depth5[2], "ask_orders": depth5[3],
                  "bid_price": depth5[4], "ask_price": depth5[5]}]

        market_depth = {
            "type" : 'Market Depth',
            "exchange_segment" : exchange_segment,
            "security_id" : sec_id,
            "LTP" : ltp,
            "depth" : depth
                   }
        return market_depth

    def process_quote(self, data):
        """Parse and process Quote Data"""
        unpack_quote = [struct.unpack('<BHBIfHIfIIIffff', data[0:50])]
        quote_data = {
            "type" : 'Quote Data',
            "exchange_segment" : unpack_quote[0][2],
            "security_id": unpack_quote[0][3],
            "LTP": "{:.2f}".format(unpack_quote[0][4]),
            "LTQ" : unpack_quote[0][5],
            "LTT" : self.utc_time(unpack_quote[0][6]),
            "avg_price" : "{:.2f}".format(unpack_quote[0][7]),
            "volume" : unpack_quote[0][8],
            "total_sell_quantity" : unpack_quote[0][9],
            "total_buy_quantity" : unpack_quote[0][10],
            "open" : "{:.2f}".format(unpack_quote[0][11]),
            "close": "{:.2f}".format(unpack_quote[0][12]),
            "high": "{:.2f}".format(unpack_quote[0][13]),
            "low": "{:.2f}".format(unpack_quote[0][14])
        }
        return quote_data
    
    def process_oi(self, data):
        """Parse and process OI Data"""
        unpack_oi = [struct.unpack('<BHBII', data[0:12])]
        oi_data = {
            "type" : 'OI Data',
            "exchange_segment" : unpack_oi[0][2],
            "security_id": unpack_oi[0][3],
            "OI": unpack_oi[0][4]
        }
        return oi_data
    
    def process_status(self, data):
        """Parse and process market status"""
        unpack_status = [struct.unpack('<BHBI', data[0:8])]
        market_status = "Markets Open"
        return market_status
    
    def server_disconnection(self, data):
        """Parse and process server disconnection error"""
        disconnection_packet = [struct.unpack('<BHBIH', data[0:10])]
        self.on_close = False
        if disconnection_packet[0][4] == 805:
            print ("Disconnected: No. of active websocket connections exceeded")
            self.on_close = True
        elif disconnection_packet[0][4] == 806:
            print ("Disconnected: Subscribe to Data APIs to continue")
            self.on_close = True
        elif disconnection_packet[0][4] == 807:
            print ("Disconnected: Access Token is expired")
            self.on_close = True
        elif disconnection_packet[0][4] == 808:
            print ("Disconnected: Invalid Client ID")
            self.on_close = True
        elif disconnection_packet[0][4] == 809:
            print ("Disconnected: Authentication Failed - check ")
            self.on_close = True
        
    async def on_connection_opened(self, websocket):
        "Callback function executed when the WebSocket connection is opened."
        await websocket.send(self.create_subscription_packet(self.instruments, subscribe=True))

        while True:
            response = await websocket.recv()
            await self.on_message_received(response)

    def pad_with_zeros(self, data, length):
        """Pads a binary data string with zeros to a specified length for server to read."""
        data = data.ljust(length, b'\0')
        return data

    def create_header(self, feed_request_code, message_length, client_id):
        """Creates header packet for the subscription packet."""
        dhan_auth = b"\0" * 50
        header = struct.pack('<bH30s50s', feed_request_code, message_length, client_id.encode('utf-8'), dhan_auth)
        return header

    def utc_time(self, epoch_time):
        """Converts EPOCH time to UTC time."""
        return datetime.utcfromtimestamp(epoch_time).strftime('%H:%M:%S')

    def create_subscription_packet(self, instruments, feed_request_code):
        """Creates the subscription packet with specified instruments and subscription code"""
        num_instruments = len(instruments)
        
        header = self.create_header(feed_request_code=self.subscription_code,
                                    message_length=83 + 4 + num_instruments * 21, 
                                    client_id=self.client_id)
        num_instruments_bytes = struct.pack('<I', num_instruments)
        instrument_info = b""
        for exchange_segment, security_id in instruments:
            instrument_info += struct.pack('<B20s', exchange_segment, security_id.encode('utf-8'))
        
        instruments = [(0, "")]
        for i in range(100 - num_instruments):
            instrument_info += struct.pack('<B20s', instruments[0][0], instruments[0][1].encode('utf-8'))

        subscription_packet = header + num_instruments_bytes + instrument_info
        return subscription_packet

    def subscribe_symbols(self, feed_request_code, symbols):
        """Function to subscribe to additional symbols."""
        unique_symbols_set = set(self.instruments)
        unique_symbols_set.update(symbols)
        self.instruments = list(unique_symbols_set)
        if self.ws and self.ws.open:
            asyncio.ensure_future(self.ws.send(self.create_subscription_packet(symbols, feed_request_code, subscribe=True)))

    def unsubscribe_symbols(self, feed_request_code, symbols):
        """Function to unsubscribe symbols from connection."""
        unique_symbols_set = set(self.instruments)
        unique_symbols_set.update(symbols)
        self.instruments = list(unique_symbols_set)
        if self.ws and self.ws.open:
            print("check it", self.instruments)
            asyncio.ensure_future(self.ws.send(self.create_subscription_packet(symbols, feed_request_code, subscribe=False)))


import pathlib
from setuptools import setup, find_packages

HERE = pathlib.Path(__file__).parent

VERSION = '1.3'
PACKAGE_NAME = 'dhanhq'
AUTHOR = 'Dhan'
AUTHOR_EMAIL = 'dhan-oss@dhan.co'
URL = ''

LICENSE = 'MIT LICENSE'
DESCRIPTION = 'The official Python client for communicating with the Dhan API'
LONG_DESCRIPTION = (HERE / "README.md").read_text()
LONG_DESC_TYPE = "text/markdown"

INSTALL_REQUIRES = ["pandas>=1.4.3",
                    "requests>=2.28.1"]
                    

setup(name=PACKAGE_NAME,
      version=VERSION,
      description=DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      long_description_content_type=LONG_DESC_TYPE,
      author=AUTHOR,
      license=LICENSE,
      author_email=AUTHOR_EMAIL,
      url=URL,
      install_requires=INSTALL_REQUIRES,
      packages=find_packages()
      )